@@ -1,6 +1,7 @@
 import os
 import sys
 import time
+import weakref
 import logging
 from contextlib import contextmanager
 import ctypes.util
@@ -41,6 +42,21 @@ def log_exception(kind):
             logger.error(msg + f" ({count})")
 
 
+def weakbind(method):
+    """Replace a bound method with a callable object that stores the `self` using a weakref."""
+    ref = weakref.ref(method.__self__)
+    class_func = method.__func__
+    del method
+
+    def proxy(*args, **kwargs):
+        self = ref()
+        if self is not None:
+            return class_func(self, *args, **kwargs)
+
+    proxy.__name__ = class_func.__name__
+    return proxy
+
+
 class WgpuCanvasInterface:
     """This is the interface that a canvas object must implement in order
     to be a valid canvas that wgpu can work with.
@@ -123,6 +139,9 @@ class WgpuCanvasBase(WgpuCanvasInterface):
         self._vsync = bool(vsync)
         self._err_hashes = {}
 
+    def __del__(self):
+        self.close()
+
     def draw_frame(self):
         """The function that gets called at each draw. You can implement
         this method in a subclass, or set it via a call to request_draw().
@@ -178,7 +197,7 @@ class WgpuCanvasBase(WgpuCanvasInterface):
 
     def close(self):
         """Close the window."""
-        raise NotImplementedError()
+        pass
 
     def is_closed(self):
         """Get whether the window is closed."""
